#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
import os.path
import subprocess
import sys
import traceback

# import interfaces
from pyams_scheduler.interfaces.ssh import ISSHCallerTask

# import packages
import paramiko
from paramiko.ssh_exception import SSHException
from pyams_scheduler.task import Task
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty


@implementer(ISSHCallerTask)
class SSHCallerTask(Task):
    """SSH caller task"""

    hostname = FieldProperty(ISSHCallerTask['hostname'])
    port = FieldProperty(ISSHCallerTask['port'])
    username = FieldProperty(ISSHCallerTask['username'])
    private_key = FieldProperty(ISSHCallerTask['private_key'])
    password = FieldProperty(ISSHCallerTask['password'])
    cmdline = FieldProperty(ISSHCallerTask['cmdline'])

    def run(self, report):
        if self.hostname:
            self._run_remote(report)
        else:
            self._run_local(report)

    def _run_remote(self, report):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(self.hostname, self.port, self.username, self.password,
                    key_filename=os.path.expanduser(self.private_key) if self.private_key else None)
        try:
            stdin, stdout, stderr = ssh.exec_command(self.cmdline)
            stdin.close()
            report.write(stdout.read().decode())
            errors = stderr.read()
            if errors:
                report.write('\n\nSome errors occurred\n===================\n')
                report.write(errors.decode())
        except SSHException:
            etype, value, tb = sys.exc_info()
            report.write('\n\nAn error occurred\n================\n')
            report.write(''.join(traceback.format_exception(etype, value, tb)))

    def _run_local(self, report):
        shell = subprocess.Popen(self.cmdline, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = shell.communicate()
        report.write(stdout.decode())
        if stderr:
            report.write('\n\nSome errors occurred\n===================\n')
            report.write(stderr.decode())
