#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'

# import standard library
from io import StringIO

# import interfaces
from pyams_form.interfaces.form import IWidgetsSuffixViewletsManager, IWidgetsPrefixViewletsManager, IFormHelp
from pyams_scheduler.interfaces import ITaskInfo, IScheduler, ITask, ICronTask, ICronTaskScheduling, IDateTask, \
    IDateTaskScheduling, ILoopTask, ILoopTaskScheduling, ITaskHistory, IBaseTaskScheduling
from pyams_skin.layer import IPyAMSLayer
from pyams_utils.interfaces import MANAGE_SYSTEM_PERMISSION
from pyams_zmi.layer import IAdminLayer
from z3c.form.interfaces import IDataExtractedEvent, DISPLAY_MODE
from z3c.table.interfaces import IValues, IColumn
from zope.traversing.interfaces import ITraversable

# import packages
from pyams_form.form import AJAXEditForm
from pyams_form.help import FormHelp
from pyams_form.schema import CloseButton
from pyams_pagelet.pagelet import pagelet_config
from pyams_skin.table import BaseTable, I18nColumn
from pyams_template.template import template_config
from pyams_utils.adapter import adapter_config, ContextAdapter, ContextRequestViewAdapter
from pyams_utils.date import format_datetime
from pyams_utils.registry import query_utility
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config, Viewlet
from pyams_zmi.form import AdminDialogAddForm, AdminDialogEditForm, AdminDialogDisplayForm
from pyramid.events import subscriber
from pyramid.url import resource_url
from pyramid.view import view_config
from z3c.form import field, button
from z3c.table.column import GetAttrColumn
from zope.interface import Invalid, Interface

from pyams_scheduler import _


@adapter_config(name='history', context=ITask, provides=ITraversable)
class TaskHistoryTraverser(ContextAdapter):
    """Task ++history++ traverser"""

    def traverse(self, name, furtherpath=None):
        return self.context.history


class TaskBaseAddForm(AdminDialogAddForm):
    """Scheduler task base add form"""

    title = _("Tasks scheduler")
    legend = _("Add URL caller task")
    label_css_class = 'control-label col-md-4'
    input_css_class = 'col-md-8'

    fields = field.Fields(ITaskInfo).omit('__parent__', '__name__')
    edit_permission = MANAGE_SYSTEM_PERMISSION
    task_factory = None

    def updateWidgets(self, prefix=None):
        super(TaskBaseAddForm, self).updateWidgets()
        self.widgets['history_duration'].value = 100
        self.widgets['history_length'].value = 100

    def create(self, data):
        return self.task_factory()

    def add(self, task):
        context = query_utility(IScheduler)
        context[task.name.lower()] = task

    def nextURL(self):
        return resource_url(self.context, self.request, 'scheduler-tasks.html')


@subscriber(IDataExtractedEvent, form_selector=TaskBaseAddForm)
def handle_new_task_data_extraction(event):
    """Handle new task form data extraction"""
    manager = query_utility(IScheduler)
    name = event.data.get('name')
    if name.lower() in manager:
        event.form.widgets.errors += (Invalid(_("Specified task name is already used!")),)


@pagelet_config(name='properties.html', context=ITask, layer=IPyAMSLayer, permission=MANAGE_SYSTEM_PERMISSION)
class TaskPropertiesEditForm(AdminDialogEditForm):
    """Scheduler task properties edit form"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("Scheduler task: {0}")).format(self.context.name)

    legend = _("Edit task properties")
    icon_css_class = 'fa fa-fw fa-clock-o'
    label_css_class = 'control-label col-md-4'
    input_css_class = 'col-md-8'

    fields = field.Fields(ITaskInfo).omit('__parent__', '__name__')
    ajax_handler = 'properties.json'
    edit_permission = MANAGE_SYSTEM_PERMISSION

    def updateWidgets(self, prefix=None):
        super(TaskPropertiesEditForm, self).updateWidgets(prefix)
        self.widgets['name'].mode = DISPLAY_MODE


@view_config(name='properties.json', context=ITask, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class SchedulerTaskPropertiesAJAXEditForm(AJAXEditForm, TaskPropertiesEditForm):
    """Scheduler task properties edit form, AJAX view"""


class TaskScheduleEditForm(AdminDialogEditForm):
    """Scheduler task base schedule edit form"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("Scheduler task: {0}")).format(self.context.name)

    legend = _("Schedule task")
    icon_css_class = 'fa fa-fw fa-calendar'
    label_css_class = 'control-label col-md-4'
    input_css_class = 'col-md-8'

    ajax_handler = 'schedule.json'
    edit_permission = MANAGE_SYSTEM_PERMISSION

    def update_content(self, content, data):
        changes = super(TaskScheduleEditForm, self).update_content(content, data)
        if changes:
            self.context.reset()
        return changes


class TaskSchedulerAJAXEditForm(AJAXEditForm):
    """Scheduler task base schedule edit form, AJAX view"""

    def get_ajax_output(self, changes):
        if 'active' in changes.get(IBaseTaskScheduling, ()):
            return {'status': 'reload'}
        else:
            return super(TaskSchedulerAJAXEditForm, self).get_ajax_output(changes)


@pagelet_config(name='schedule.html', context=ICronTask, layer=IPyAMSLayer, permission=MANAGE_SYSTEM_PERMISSION)
class CronTaskScheduleEditForm(TaskScheduleEditForm):
    """Cron-style task schedule edit form"""

    fields = field.Fields(ICronTaskScheduling)


@view_config(name='schedule.json', context=ICronTask, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class CronTaskScheduleAJAXEditForm(TaskSchedulerAJAXEditForm, CronTaskScheduleEditForm):
    """Cron-style task schedule edit form, AJAX view"""


@pagelet_config(name='schedule.html', context=IDateTask, layer=IPyAMSLayer, permission=MANAGE_SYSTEM_PERMISSION)
class DateTaskScheduleEditForm(TaskScheduleEditForm):
    """Date-style task schedule edit form"""

    fields = field.Fields(IDateTaskScheduling)


@view_config(name='schedule.json', context=IDateTask, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class DateTaskScheduleAJAXEditForm(TaskSchedulerAJAXEditForm, DateTaskScheduleEditForm):
    """Date-style task schedule edit form, AJAX view"""


@pagelet_config(name='schedule.html', context=ILoopTask, layer=IPyAMSLayer, permission=MANAGE_SYSTEM_PERMISSION)
class LoopTaskScheduleEditForm(TaskScheduleEditForm):
    """Loop-style task schedule edit form"""

    fields = field.Fields(ILoopTaskScheduling)


@view_config(name='schedule.json', context=ILoopTask, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class LoopTaskScheduleAJAXEditForm(TaskSchedulerAJAXEditForm, LoopTaskScheduleEditForm):
    """Loop-style task schedule edit form, AJAX view"""


class ITaskRunnerButtons(Interface):
    """Task runner buttons"""

    close = CloseButton(name='close', title=_("Close"))
    debug = button.Button(name='debug', title=_("Run in debug mode"))
    execute = button.Button(name='execute', title=_("Run in normal mode"))


@pagelet_config(name='run.html', context=ITask, layer=IPyAMSLayer, permission=MANAGE_SYSTEM_PERMISSION)
class TaskRunForm(AdminDialogEditForm):
    """Task runner form"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("Scheduler task: {0}")).format(self.context.name)

    dialog_class = 'modal-large'
    legend = _("Execute task")
    fields = field.Fields(Interface)
    buttons = button.Buttons(ITaskRunnerButtons)

    ajax_handler = 'run.json'

    def updateActions(self):
        super(TaskRunForm, self).updateActions()
        if 'debug' in self.actions:
            self.actions['debug'].addClass('btn-info')
        if 'execute' in self.actions:
            self.actions['execute'].addClass('btn-primary')

    def applyChanges(self, data):
        task = self.getContent()
        if self.actions['execute'].name in self.request.params:
            task.launch()
        else:
            report = StringIO()
            task.run(report)
            return report.getvalue()


@view_config(name='run.json', context=ITask, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class TaskRunAJAXForm(AJAXEditForm, TaskRunForm):
    """Task runner form, AJAX view"""

    def get_ajax_output(self, changes):
        translate = self.request.localizer.translate
        if self.actions['execute'].name in self.request.params:
            return {'status': 'success',
                    'message': translate(_('Task scheduled in normal mode will start in 5 seconds...'))}
        else:
            return {'status': 'message',
                    'content': {'raw': True,
                                'text': changes,
                                'target': '#task-debug-report'},
                    'message': translate(_('Task run in debug mode. Please check your console...')),
                    'close_form': False}


@adapter_config(context=(ITask, IAdminLayer, TaskRunForm), provides=IFormHelp)
class TaskRunFormHelpAdapter(FormHelp):
    """Task run form help adapter"""

    status = 'warning'
    header = _("Executing and debugging task")
    message = _("""You can choose to execute the task in 'normal' mode or in 'debug' mode.

In normal mode, the task is scheduled in a standard way and run in a background process (after 5 seconds).

In debug mode, the task is run in the context of the main application process; the goal of this mode
is to allow a developer to insert breakpoints.

**WARNING**: in both mode, the task will be executed even if it's disabled in it's scheduling settings!""")
    message_format = 'rest'


@viewlet_config(name='task-debug-report', view=TaskRunForm, layer=IAdminLayer,
                manager=IWidgetsPrefixViewletsManager)
@template_config(template='templates/task-debug-report.pt')
class TaskDebugReportViewlet(Viewlet):
    """Task debug report viewlet"""


@pagelet_config(name='history.html', context=ITask, layer=IPyAMSLayer, permission=MANAGE_SYSTEM_PERMISSION)
class TaskHistoryDisplayForm(AdminDialogDisplayForm):
    """Task history display form"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("Scheduler task: {0}")).format(self.context.name)

    legend = _("Task history")
    dialog_class = 'modal-max'
    icon_css_class = 'fa fa-fw fa-history'

    fields = field.Fields(Interface)


STATUS_CLASS = {'OK': 'success',
                'Warning': 'warning',
                'Error': 'danger',
                'Empty': 'info'}


class TaskHistoryItemsTable(BaseTable):
    """Task history items table"""

    title = _("Task history")
    cssClasses = {'table': 'table table-bordered table-striped table-hover table-tight datatable'}
    sortOn = None

    @property
    def data_attributes(self):
        return {'table': {'data-ams-datatable-global-filter': 'false',
                          'data-ams-datatable-info': 'false',
                          'data-ams-datatable-sort': 'false',
                          'data-ams-datatable-sdom': "t<'dt-row dt-bottom-row'<'text-right'p>>",
                          'data-ams-datatable-display-length': '20',
                          'data-ams-datatable-pagination-type': 'bootstrap_prevnext'},
                'tr': {'data-ams-url': lambda x, col: absolute_url(x, self.request, 'info.json'),
                       'data-ams-target': '#task-history-report'}}

    def getCSSHighlightClass(self, column, item, cssClass):
        return STATUS_CLASS[item.status]


@adapter_config(name='name', context=(Interface, IAdminLayer, TaskHistoryItemsTable), provides=IColumn)
class TaskHistoryDateColumn(I18nColumn, GetAttrColumn):
    """Task history date column"""

    _header = _("Date")
    attrName = 'date'

    def renderCell(self, item):
        return format_datetime(item.date, request=self.request)


@adapter_config(context=(ITask, IAdminLayer, TaskHistoryItemsTable), provides=IValues)
class TaskHistoryValuesAdapter(ContextRequestViewAdapter):
    """Task history values adapter"""

    @property
    def values(self):
        return sorted(self.context.history.values(),
                      key=lambda x: x.date,
                      reverse=True)


@viewlet_config(name='task-history', view=TaskHistoryDisplayForm, layer=IAdminLayer,
                manager=IWidgetsSuffixViewletsManager)
@template_config(template='templates/task-history.pt')
class TaskHistoryViewlet(Viewlet):
    """Task history viewlet"""

    table = TaskHistoryItemsTable

    def __init__(self, context, request, view, manager):
        super(TaskHistoryViewlet, self).__init__(context, request, view, manager)
        self.table = TaskHistoryItemsTable(context, request)

    def update(self):
        super(TaskHistoryViewlet, self).update()
        self.table.update()


@view_config(name='info.json', context=ITaskHistory, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
def TaskHistoryInfoView(request):
    return {'status': 'success',
            'close_form': False,
            'content': {'raw': True,
                        'text': request.context.report,
                        'target': '#task-history-report'}}
